#include <stdio.h>
#include <stdarg.h>

#include "printing.h"

#ifdef HAVE_WRITE
    #include <unistd.h>
#endif

#ifdef HAVE_GETTIMEOFDAY
    #include <sys/time.h>
#else
    #include <time.h>
#endif

static int thread_indent[MAX_THREADS];

void print(const char *format, ...) {
    char buffer[MAX_MESSAGE_LENGTH];
    va_list arg;
    va_start(arg, format);

    strcpy(buffer, "probe: ");  /* 7 chars */

#ifdef HAVE_SNPRINTF
    vsnprintf(buffer+7, MAX_MESSAGE_LENGTH, format, arg);
#else
    vsprintf(buffer+7, format, arg);
#endif

    /* We format the string into a buffer and then call write directly,
       because this has the least chance of interleaving output due to
       threading. The output might still be interleaved but this is about
       the best we can do without using locks and possibly affecting the
       behaviour of the program drastically, or implementing a complex
       non-locking data structure.
    */
    write(STDERR_FILENO, buffer, strlen(buffer));

    va_end(arg);
}

void print_callback(const char *source, int32_t thread,
    const char *format, ...) {

    char buffer[MAX_MESSAGE_LENGTH];
    int length = 0;
    int indent = thread_indent[thread];
    va_list arg;
    va_start(arg, format);

    strcpy(buffer, "probe: ");  /* 7 chars */
    length += 7;

#ifdef HAVE_SNPRINTF
    length += snprintf(buffer + length, MAX_MESSAGE_LENGTH - length,
#else
    length += sprintf(buffer + length,
#endif
        "[%8.6f %s thd %2d] ", current_time(), source, thread);

    while(indent) {
        buffer[length ++] = ' ';
        buffer[length ++] = ' ';
        indent --;
    }

#ifdef HAVE_VSNPRINTF
    vsnprintf(buffer + length, MAX_MESSAGE_LENGTH - length, format, arg);
#else
    vsprintf(buffer + length, format, arg);
#endif

    /* We format the string into a buffer and then call write directly,
       because this has the least chance of interleaving output due to
       threading. The output might still be interleaved but this is about
       the best we can do without using locks and possibly affecting the
       behaviour of the program drastically, or implementing a complex
       non-locking data structure.
    */
    write(STDERR_FILENO, buffer, strlen(buffer));

    va_end(arg);
}

void add_indent(int32_t thread) {
    thread_indent[thread] ++;
}

void subtract_indent(int32_t thread) {
    if(thread_indent[thread]) {
        thread_indent[thread] --;
    }
}

#ifdef HAVE_GETTIMEOFDAY
static double initial_time;

void init_timing(void) {
    initial_time = current_time();
}

double current_time(void) {
    struct timeval tv;
    double seconds;

    gettimeofday(&tv, NULL);

    seconds = tv.tv_sec;
    seconds += tv.tv_usec / 1000000.0;
    return seconds - initial_time;
}
#else
static clock_t initial_time;

void init_timing(void) {
    initial_time = clock();
}

double current_time(void) {
    clock_t now = clock();

    return (double)(now - initial_time) / CLOCKS_PER_SEC;
}
#endif

